# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run this to generate BIG-bench SeqIO Task catalog.

The catalog contains details of BIG-bench JSON tasks and corresponding canonical
SeqIO tasks available for use.
"""
from absl import app
from bigbench.bbseqio.tools import utils
from bigbench.bbseqio import bigbench_bridge as bb
from bigbench.bbseqio import bigbench_json_paths
from bigbench.bbseqio import bigbench_lite as bb_lite
from bigbench.bbseqio import tasks  # pylint:disable=unused-import
import bigbench.bbseqio.utils as bbseqio_utils
import seqio


def make_collapsible(heading: str, text: str) -> str:
  return f"<details><summary>{heading}</summary><br>{text}</details>"


def get_seqio_tasks_catalog_lines(seqio_task_names, seqio_mixture_names):
  """Generates a BIG-bench SeqIO Task catalog."""
  bb_task_names = sorted(
      bigbench_json_paths.get_task_names(), key=lambda x: x.lower())

  # Group tasks with and without subtasks.
  bb_task_names = sorted(bb_task_names, key=bigbench_json_paths.has_subtasks)

  lines = []
  lines.append(f"{utils.get_file_header()}\n\n")
  # BIG-bench Lite Mixtures.
  lines.append("## BIG-bench Lite SeqIO Mixtures\n\n")
  lines.append("SeqIO Mixture | BIG-bench Tasks")
  lines.append("----|----")
  for bb_lite_name, bb_lite_task_names in bb_lite.BIGBENCH_LITE_TASK_NAMES.items(
  ):
    bb_lite_seqio_mixtures = sorted(
        [m for m in seqio.MixtureRegistry.names() if f":{bb_lite_name}" in m])
    bb_lite_seqio_mixtures = [f"`{m}`" for m in bb_lite_seqio_mixtures]
    bb_lite_seqio_mixtures = "<br>".join(bb_lite_seqio_mixtures)
    bb_lite_task_names = sorted(bb_lite_task_names, key=lambda x: x.lower())
    bb_lite_task_urls = [
        utils.get_bb_task_url(bb_task_name)
        for bb_task_name in bb_lite_task_names
    ]
    bb_lite_task_names = [
        f"[{name}]({url})"
        for name, url in zip(bb_lite_task_names, bb_lite_task_urls)
    ]
    bb_lite_task_names = "<br>".join(bb_lite_task_names)
    lines.append(f"{bb_lite_seqio_mixtures} | {bb_lite_task_names}")
  lines.append("\n\n")

  # All BIG-bench Tasks.
  lines.append("## All BIG-bench SeqIO Tasks\n\n")
  lines.append("BIG-bench Task | SeqIO Tasks | Subtasks | Description | Type")
  lines.append("----|----|----|----|----")
  for bb_task_name in bb_task_names:
    # BB Task
    bb_task_url = utils.get_bb_task_url(bb_task_name)
    row = f"[{bb_task_name}]({bb_task_url})"

    # SeqIO Tasks/Mixtures
    formatted_bb_task_name = bbseqio_utils.format_name_for_seqio(bb_task_name)
    seqio_tasks = sorted([
        f"`{t}`" for t in seqio_task_names if f":{formatted_bb_task_name}." in t
    ])
    seqio_mixtures = sorted([
        f"`{m}`" for m in seqio_mixture_names
        if f":{formatted_bb_task_name}." in m and ".mix." in m
    ])
    to_show = seqio_mixtures if seqio_mixtures else seqio_tasks
    heading = "Tasks"
    column = make_collapsible(heading, "<br>".join(to_show))
    row = f"{row} | {column} "

    # Subtasks
    subtask_names = "-"
    if bigbench_json_paths.has_subtasks(bb_task_name):
      subtask_names = sorted([
          bbseqio_utils.format_name_for_seqio(subtask_name.split(":")[-1]) for
          subtask_name in bigbench_json_paths.get_subtask_names(bb_task_name)
      ])
      subtask_names = make_collapsible("Subtasks", "<br>".join(subtask_names))
    row = f"{row} | {subtask_names}"

    # Description
    description = bigbench_json_paths.get_task_description(bb_task_name)
    column = make_collapsible("Description", description)
    row = f"{row} | {column}"

    # Types
    task_types = []
    if bigbench_json_paths.has_subtasks(bb_task_name):
      for subtask_name in bigbench_json_paths.get_subtask_names(bb_task_name):
        task_types.extend(bigbench_json_paths.get_task_types(subtask_name))
    else:
      task_types = bigbench_json_paths.get_task_types(bb_task_name)
    task_types = set(task_types)
    task_types = sorted([
        bb.BigBenchTaskType.to_full_str(task_type) for task_type in task_types
    ])
    task_types = ", ".join(task_types)
    row = f"{row} | {task_types}"

    # Append row to table
    lines.append(row)

  return lines


def main(_) -> None:
  seqio_task_names = [t for t in seqio.TaskRegistry.names() if "bigbench:" in t]
  seqio_mixture_names = [
      m for m in seqio.MixtureRegistry.names() if "bigbench:" in m
  ]
  lines = get_seqio_tasks_catalog_lines(seqio_task_names, seqio_mixture_names)
  utils.write_to_catalog_file(lines)


if __name__ == "__main__":
  app.run(main)
